On 08/01/21 07:49, Chenyi Qiang wrote:
To avoid breaking the CPUs without bus lock detection, activate the
DR6_BUS_LOCK bit (bit 11) conditionally in DR6_FIXED_1 bits.

The set/clear of DR6_BUS_LOCK is similar to the DR6_RTM in DR6
register. The processor clears DR6_BUS_LOCK when bus lock debug
exception is generated. (For all other #DB the processor sets this bit
to 1.) Software #DB handler should set this bit before returning to the
interrupted task.

For VM exit caused by debug exception, bit 11 of the exit qualification
is set to indicate that a bus lock debug exception condition was
detected. The VMM should emulate the exception by clearing bit 11 of the
guest DR6.

Please rename DR6_INIT to DR6_ACTIVE_LOW, and then a lot of changes become simpler:

-               dr6 |= DR6_BD | DR6_RTM;
+               dr6 |= DR6_BD | DR6_RTM | DR6_BUS_LOCK;

dr6 |= DR6_BD | DR6_ACTIVE_LOW;

                ctxt->ops->set_dr(ctxt, 6, dr6);
                return emulate_db(ctxt);
        }
diff --git a/arch/x86/kvm/svm/svm.c b/arch/x86/kvm/svm/svm.c
index cce0143a6f80..3d8a0e30314f 100644
--- a/arch/x86/kvm/svm/svm.c
+++ b/arch/x86/kvm/svm/svm.c
@@ -1860,7 +1860,7 @@ static void svm_sync_dirty_debug_regs(struct kvm_vcpu 
*vcpu)
        get_debugreg(vcpu->arch.db[2], 2);
        get_debugreg(vcpu->arch.db[3], 3);
        /*
-        * We cannot reset svm->vmcb->save.dr6 to DR6_FIXED_1|DR6_RTM here,
+        * We cannot reset svm->vmcb->save.dr6 to 
DR6_FIXED_1|DR6_RTM|DR6_BUS_LOCK here,

We cannot reset svm->vmcb->save.dr6 to DR6_ACTIVE_LOW

         * because db_interception might need it.  We can do it before vmentry.
         */
        vcpu->arch.dr6 = svm->vmcb->save.dr6;
@@ -1911,7 +1911,7 @@ static int db_interception(struct vcpu_svm *svm)
        if (!(svm->vcpu.guest_debug &
              (KVM_GUESTDBG_SINGLESTEP | KVM_GUESTDBG_USE_HW_BP)) &&
                !svm->nmi_singlestep) {
-               u32 payload = (svm->vmcb->save.dr6 ^ DR6_RTM) & ~DR6_FIXED_1;
+               u32 payload = (svm->vmcb->save.dr6 ^ (DR6_RTM|DR6_BUS_LOCK)) & 
~DR6_FIXED_1;

u32 payload = svm->vmcb->save.dr6 ^ DR6_ACTIVE_LOW;

                kvm_queue_exception_p(&svm->vcpu, DB_VECTOR, payload);
                return 1;
        }
@@ -3778,7 +3778,7 @@ static __no_kcsan fastpath_t svm_vcpu_run(struct kvm_vcpu 
*vcpu)
        if (unlikely(svm->vcpu.arch.switch_db_regs & KVM_DEBUGREG_WONT_EXIT))
                svm_set_dr6(svm, vcpu->arch.dr6);
        else
-               svm_set_dr6(svm, DR6_FIXED_1 | DR6_RTM);
+               svm_set_dr6(svm, DR6_FIXED_1 | DR6_RTM | DR6_BUS_LOCK);

svm_set_dr6(svm, DR6_ACTIVE_LOW);

clgi();
        kvm_load_guest_xsave_state(vcpu);
diff --git a/arch/x86/kvm/vmx/nested.c b/arch/x86/kvm/vmx/nested.c
index e2f26564a12d..c5d71a9b3729 100644
--- a/arch/x86/kvm/vmx/nested.c
+++ b/arch/x86/kvm/vmx/nested.c
@@ -412,7 +412,7 @@ static int nested_vmx_check_exception(struct kvm_vcpu 
*vcpu, unsigned long *exit
                        if (!has_payload) {
                                payload = vcpu->arch.dr6;
                                payload &= ~(DR6_FIXED_1 | DR6_BT);
-                               payload ^= DR6_RTM;
+                               payload ^= DR6_RTM | DR6_BUS_LOCK;

payload &= ~DR6_BT;
payload ^= DR6_ACTIVE_LOW;

                        }
                        *exit_qual = payload;
                } else
diff --git a/arch/x86/kvm/x86.c b/arch/x86/kvm/x86.c
index 3f7c1fc7a3ce..06de2b9e57f3 100644
--- a/arch/x86/kvm/x86.c
+++ b/arch/x86/kvm/x86.c
@@ -483,19 +483,20 @@ void kvm_deliver_exception_payload(struct kvm_vcpu *vcpu)
                 */
                vcpu->arch.dr6 &= ~DR_TRAP_BITS;
                /*
                 * DR6.RTM is set by all #DB exceptions that don't clear it.
                 */
-               vcpu->arch.dr6 |= DR6_RTM;
+               vcpu->arch.dr6 |= DR6_RTM | DR6_BUS_LOCK;
                vcpu->arch.dr6 |= payload;
                /*
                 * Bit 16 should be set in the payload whenever the #DB
                 * exception should clear DR6.RTM. This makes the payload
                 * compatible with the pending debug exceptions under VMX.
                 * Though not currently documented in the SDM, this also
                 * makes the payload compatible with the exit qualification
                 * for #DB exceptions under VMX.
                 */
-               vcpu->arch.dr6 ^= payload & DR6_RTM;
+               vcpu->arch.dr6 ^= payload & (DR6_RTM | DR6_BUS_LOCK);

vcpu->arch.dr6 &= ~DR_TRAP_BITS;
vcpu->arch.dr6 |= DR6_ACTIVE_LOW;
vcpu->arch.dr6 |= payload;
vcpu->arch.dr6 ^= payload & DR6_ACTIVE_LOW;

(with comments :))

and so on.

Thanks!

Paolo

/*
                 * The #DB payload is defined as compatible with the 'pending
@@ -1126,6 +1127,9 @@ static u64 kvm_dr6_fixed(struct kvm_vcpu *vcpu)
if (!guest_cpuid_has(vcpu, X86_FEATURE_RTM))
                fixed |= DR6_RTM;
+
+       if (!guest_cpuid_has(vcpu, X86_FEATURE_BUS_LOCK_DETECT))
+               fixed |= DR6_BUS_LOCK;
        return fixed;
  }
@@ -7197,7 +7201,8 @@ static int kvm_vcpu_do_singlestep(struct kvm_vcpu *vcpu)
        struct kvm_run *kvm_run = vcpu->run;
if (vcpu->guest_debug & KVM_GUESTDBG_SINGLESTEP) {
-               kvm_run->debug.arch.dr6 = DR6_BS | DR6_FIXED_1 | DR6_RTM;
+               kvm_run->debug.arch.dr6 = DR6_BS | DR6_FIXED_1 | DR6_RTM |
+                                         DR6_BUS_LOCK;
                kvm_run->debug.arch.pc = kvm_get_linear_rip(vcpu);
                kvm_run->debug.arch.exception = DB_VECTOR;
                kvm_run->exit_reason = KVM_EXIT_DEBUG;
@@ -7241,7 +7246,8 @@ static bool kvm_vcpu_check_breakpoint(struct kvm_vcpu 
*vcpu, int *r)
                                           vcpu->arch.eff_db);
if (dr6 != 0) {
-                       kvm_run->debug.arch.dr6 = dr6 | DR6_FIXED_1 | DR6_RTM;
+                       kvm_run->debug.arch.dr6 = dr6 | DR6_FIXED_1 | DR6_RTM |
+                                                 DR6_BUS_LOCK;
                        kvm_run->debug.arch.pc = eip;
                        kvm_run->debug.arch.exception = DB_VECTOR;
                        kvm_run->exit_reason = KVM_EXIT_DEBUG;
@@ -483,19 +483,20 @@ void kvm_deliver_exception_payload(struct kvm_vcpu *vcpu)
                vcpu->arch.dr6 &= ~DR_TRAP_BITS;
                /*
-                * DR6.RTM is set by all #DB exceptions that don't clear it.
+                * DR6.RTM and DR6.BUS_LOCK are set by all #DB exceptions
+                * that don't clear it.
                 */
@@ -483,19 +483,20 @@ void kvm_deliver_exception_payload(struct kvm_vcpu *vcpu)
                vcpu->arch.dr6 |= payload;
                /*
-                * Bit 16 should be set in the payload whenever the #DB
-                * exception should clear DR6.RTM. This makes the payload
-                * compatible with the pending debug exceptions under VMX.
-                * Though not currently documented in the SDM, this also
-                * makes the payload compatible with the exit qualification
-                * for #DB exceptions under VMX.
+                * Bit 16/Bit 11 should be set in the payload whenever
+                * the #DB exception should clear DR6.RTM/DR6.BUS_LOCK.
+                * This makes the payload compatible with the pending debug
+                * exceptions under VMX. Though not currently documented in
+                * the SDM, this also makes the payload compatible with the
+                * exit qualification for #DB exceptions under VMX.
                 */


Reply via email to